@@ -4,7 +4,7 @@
 #
 # Copyright (C) 2004 Robert Kaye
 # Copyright (C) 2006-2009, 2011-2013, 2017 Lukáš Lalinský
-# Copyright (C) 2007-2011, 2015, 2018-2022 Philipp Wolfer
+# Copyright (C) 2007-2011, 2015, 2018-2023 Philipp Wolfer
 # Copyright (C) 2008 Gary van der Merwe
 # Copyright (C) 2008-2009 Nikolai Prokoschenko
 # Copyright (C) 2009 Carlin Mangar
@@ -25,8 +25,8 @@
 # Copyright (C) 2019 Joel Lintunen
 # Copyright (C) 2020 Ray Bouchard
 # Copyright (C) 2020-2021 Gabriel Ferreira
-# Copyright (C) 2021 Bob Swift
 # Copyright (C) 2021 Petit Minion
+# Copyright (C) 2021, 2023 Bob Swift
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -54,6 +54,9 @@ import os
 import os.path
 import re
 import shutil
+import time
+
+from mutagen._util import MutagenError
 
 from PyQt5 import QtCore
 
@@ -62,6 +65,7 @@ from picard import (
     log,
 )
 from picard.config import get_config
+from picard.const import DEFAULT_TIME_FORMAT
 from picard.const.sys import (
     IS_MACOS,
     IS_WIN,
@@ -94,7 +98,10 @@ from picard.util.filenaming import (
 )
 from picard.util.preservedtags import PreservedTags
 from picard.util.scripttofilename import script_to_filename_with_metadata
-from picard.util.tags import PRESERVED_TAGS
+from picard.util.tags import (
+    CALCULATED_TAGS,
+    PRESERVED_TAGS,
+)
 
 from picard.ui.item import Item
 
@@ -104,6 +111,7 @@ class FileErrorType(Enum):
     UNKNOWN = auto()
     NOTFOUND = auto()
     NOACCESS = auto()
+    PARSER = auto()
 
 
 class File(QtCore.QObject, Item):
@@ -203,13 +211,16 @@ class File(QtCore.QObject, Item):
 
     def _set_error(self, error):
         self.state = File.ERROR
-        self.error_append(str(error))
-        if any_exception_isinstance(error, FileNotFoundError):
+        if any_exception_isinstance(error, MutagenError):
+            self.error_type = FileErrorType.PARSER
+            self.error_append(_('The file failed to parse, either the file is damaged or has an unsupported file format.'))
+        elif any_exception_isinstance(error, FileNotFoundError):
             self.error_type = FileErrorType.NOTFOUND
         elif any_exception_isinstance(error, PermissionError):
             self.error_type = FileErrorType.NOACCESS
         else:
             self.error_type = FileErrorType.UNKNOWN
+        self.error_append(str(error))
 
     def load(self, callback):
         thread.run_task(
@@ -249,8 +260,8 @@ class File(QtCore.QObject, Item):
 
             if alternative_file:
                 # Do not retry reloading exactly the same file format
-                if type(alternative_file) != type(self):  # pylint: disable=unidiomatic-typecheck
-                    log.debug('Loading %r failed, retrying as %r' % (self, alternative_file))
+                if type(alternative_file) != type(self):  # pylint: disable=unidiomatic-typecheck # noqa: E721
+                    log.debug('Loading %r failed, retrying as %r', self, alternative_file)
                     self.remove()
                     alternative_file.load(callback)
                     return
@@ -275,7 +286,6 @@ class File(QtCore.QObject, Item):
             if fingerprints:
                 self.set_acoustid_fingerprint(fingerprints[0])
         run_file_post_load_processors(self)
-        self.update()
         callback(self)
 
     def _copy_loaded_metadata(self, metadata, postprocessors=None):
@@ -298,9 +308,14 @@ class File(QtCore.QObject, Item):
             to_metadata[info] = from_metadata[info]
 
     def copy_metadata(self, metadata, preserve_deleted=True):
-        acoustid = self.metadata["acoustid_id"]
         saved_metadata = {}
 
+        # Keep current value for special tags that got calculated from audio content
+        for tag in CALCULATED_TAGS:
+            if tag not in metadata.deleted_tags and self.metadata[tag]:
+                saved_metadata[tag] = self.metadata[tag]
+
+        # Keep original values of preserved tags
         preserved_tags = PreservedTags()
         for tag, values in self.orig_metadata.rawitems():
             if tag in preserved_tags or tag in PRESERVED_TAGS:
@@ -314,8 +329,6 @@ class File(QtCore.QObject, Item):
                 del self.metadata[tag]
         self.metadata.update(saved_metadata)
 
-        if acoustid and "acoustid_id" not in metadata.deleted_tags:
-            self.metadata["acoustid_id"] = acoustid
         if images_changed:
             self.metadata_images_changed.emit()
 
@@ -683,10 +696,10 @@ class File(QtCore.QObject, Item):
         for name in set(self.metadata) | set(self.orig_metadata):
             if name.startswith('~'):
                 continue
-            if not self.supports_tag(name):
-                continue
             if name in ignored_tags:
                 continue
+            if not self.supports_tag(name):
+                continue
             yield name
 
     def update(self, signal=True):
@@ -769,6 +782,17 @@ class File(QtCore.QObject, Item):
         metadata['~filename'] = filename
         metadata['~extension'] = extension.lower()[1:]
 
+        try:
+            created = os.path.getctime(self.filename)
+            created_timestamp = time.strftime(DEFAULT_TIME_FORMAT, time.localtime(created))
+            metadata['~file_created_timestamp'] = created_timestamp
+
+            modified = os.path.getmtime(self.filename)
+            modified_timestamp = time.strftime(DEFAULT_TIME_FORMAT, time.localtime(modified))
+            metadata['~file_modified_timestamp'] = modified_timestamp
+        except OSError as ex:
+            log.error(f"File Timestamps Error: {ex}")
+
     @property
     def state(self):
         """Current state of the File object"""
